(* 
 * Managing Isabelle's concrete syntax
 * $Id: syn_mng.sml 1161 2009-04-21 08:21:45Z maxim $
 *)


signature SYN_MNG =
sig
 val string_of_mixfix : mixfix -> string

 val add_type_mixfix : (string * mixfix) -> unit
 val add_mixfix : (string * mixfix) -> unit
 val add_nonterm : (string * bstring) -> unit
 val add_syn_decl : (string * (string * bool * bstring * Term.typ * mixfix)) -> unit
 val add_trrule : (string * ((xstring * string) Syntax.trrule)) -> unit
 val add_parse_ast_trans : (string * (bool * (string * Position.T))) -> unit
 val add_parse_trans : (string * (bool * (string * Position.T))) -> unit 
 val add_print_trans : (string * (bool * (string * Position.T))) -> unit
 val add_print_ast_trans : (string * (bool * (string * Position.T))) -> unit
 val add_token_trans : (string * (string * Position.T)) -> unit

 val get_type_mixfix : string -> mixfix option
 val get_mixfix : string -> mixfix option
 val get_nonterms : string -> bstring list 
 val get_syn_decls : string -> (string * bool * bstring * Term.typ * mixfix) list  
 val get_trrules : string -> ((xstring * string) Syntax.trrule) list
 val get_parse_ast_trans : string -> (bool * (string * Position.T)) list
 val get_parse_trans : string -> (bool * (string * Position.T)) list 
 val get_print_trans : string -> (bool * (string * Position.T)) list
 val get_print_ast_trans : string -> (bool * (string * Position.T)) list
 val get_token_trans : string -> (string * Position.T) list

 val upd_rem_tabs : string -> unit
 val is_transl1 : string -> theory -> bool
 val is_transl2 : string -> theory -> bool
 val add_transl1 : string -> string -> unit
 val add_transl2 : string -> string -> unit

end

structure Syn_mng : SYN_MNG =
struct

fun prt_verb s = if false then writeln s else ()

(* Tables *)
 
val type_mixfix_tab = ref Symtab.empty : (mixfix Symtab.table) ref
val mixfix_tab = ref Symtab.empty : (mixfix Symtab.table) ref
val nonterm_tab = ref Symtab.empty : ((bstring list) Symtab.table) ref
val syntax_tab = ref Symtab.empty : (((string * bool * bstring * Term.typ * mixfix) list) Symtab.table) ref

val transl_tab = ref Symtab.empty : ((string list) Symtab.table) ref
val transl_ops_tab = ref Symtab.empty : ((string list) Symtab.table) ref

val parse_ast_tab   = ref Symtab.empty : (((bool * (string * Position.T)) list) Symtab.table) ref
val parse_tab       = ref Symtab.empty : (((bool * (string * Position.T)) list) Symtab.table) ref
val print_tab       = ref Symtab.empty : (((bool * (string * Position.T)) list) Symtab.table) ref
val print_ast_tab   = ref Symtab.empty : (((bool * (string * Position.T)) list) Symtab.table) ref
val token_tab       = ref Symtab.empty : (((string * Position.T) list) Symtab.table) ref

val abbr_tab = ref Symtab.empty : 
(((Syntax.mode * (string * string option * mixfix) option * string) list) Symtab.table) ref 



(* ********************************************************* *)


fun add_transl t idt thynm = 
  let val tab = !t 
      val tl = case (Symtab.lookup tab idt) of
                SOME l => l
              | NONE => []
      val tab' = Symtab.update (idt, thynm::tl) tab
  in
   t := tab'
  end

fun is_transl t idt thy =
  let val thynm = Aux.name_of_thy thy
      val tab = !t
      val tl = case (Symtab.lookup tab idt) of
                SOME l => l
               | NONE => []
      val thy_anc = thynm::(map Aux.name_of_thy (Theory.ancestors_of thy))
      val i = thy_anc inter tl
  in
   not (i = [])
  end   

val add_transl1 = add_transl transl_tab
val add_transl2 = add_transl transl_ops_tab
val is_transl1 = is_transl transl_tab
val is_transl2 = is_transl transl_ops_tab

fun upd_rem_tab t thynm =
  let val tab = !t
      val keys = Symtab.keys tab
      fun rem [] = []
        | rem (x::xs) = (if (x = thynm) then rem xs else (x::(rem xs)))
      fun chk t [] = t
        | chk t (x::xs) = (case (Symtab.lookup tab x) of
                            NONE => error ("Unexpected error in upd_rem_tab")
                          | SOME i => (let val i' = rem i
                                           val t' = Symtab.update (x, i') t
                                       in
                                        chk t' xs
                                       end))
      val tab' = chk tab keys
  in
   t := tab'
  end

fun upd_rem_tabs thynm = 
 let val _ = upd_rem_tab transl_tab thynm
 in
 upd_rem_tab transl_ops_tab thynm
 end
                                           

(* ************* Mixfix ******************* *)

val parens = enclose "(" ")";
val brackets = enclose "[" "]";
val spaces = space_implode " ";

fun string_of_mixfix NoSyn = ""
  | string_of_mixfix (Mixfix (sy, ps, p)) =
      parens (spaces [quote sy, brackets (commas (map string_of_int ps)), string_of_int p])
  | string_of_mixfix (Delimfix sy) = parens (quote sy)

  | string_of_mixfix (InfixName (sy, p)) = parens (spaces ["infix", quote sy, string_of_int p])
  | string_of_mixfix (InfixlName (sy, p)) = parens (spaces ["infixl", quote sy, string_of_int p])
  | string_of_mixfix (InfixrName (sy, p)) = parens (spaces ["infixl", quote sy, string_of_int p])
  | string_of_mixfix (Infix p) = parens (spaces ["infix", string_of_int p])
  | string_of_mixfix (Infixl p) = parens (spaces ["infixl", string_of_int p])
  | string_of_mixfix (Infixr p) = parens (spaces ["infixr", string_of_int p])
  | string_of_mixfix (Binder (sy, p1, p2)) =
      parens (spaces ["binder", brackets (string_of_int p1), string_of_int p2]);



 fun add_type_mixfix (opname, mxfx) = 
  let val (ref tab) = type_mixfix_tab
      val tab' = Symtab.update (opname, mxfx) tab
  in
   type_mixfix_tab := tab'
  end 

 fun get_type_mixfix opname =
  let val (ref tab) = type_mixfix_tab
  in
   Symtab.lookup tab opname
  end  

 fun add_mixfix (opname, mxfx) = 
  let val (ref tab) = mixfix_tab
      val tab' = Symtab.update (opname, mxfx) tab
      val _ = prt_verb ("Mixfix for " ^ opname ^ " added.")
  in
   mixfix_tab := tab'
  end 

 fun get_mixfix opname =
  let val (ref tab) = mixfix_tab
  in
   Symtab.lookup tab opname
  end  

(* *********** Nonterminals ************ *)



 fun add_nonterm (thyname, nt) = 
  let val (ref tab) = nonterm_tab
      val nts = (case (Symtab.lookup tab thyname) of
                  SOME l => l
                 | NONE => [])
      val tab' = Symtab.update (thyname, nts @ [(nt)]) tab
  in
   nonterm_tab := tab'
  end 

 fun get_nonterms thyname =
  let val (ref tab) = nonterm_tab
  in
   case (Symtab.lookup tab thyname) of
     SOME nts => nts
   | NONE => []
  end  



(* *********** Syntax ************ *)


 fun add_syn_decl (thyname, decl) = 
  let val (ref tab) = syntax_tab
      val decls = (case (Symtab.lookup tab thyname) of
                    SOME l => l
                   | NONE => [])
      fun qlf (s, b, nm, t, m) = (s, b, thyname ^ "." ^ nm, t, m)
      val tab' = Symtab.update (thyname, decls @ [decl]) tab
  in
   syntax_tab := tab'
  end 

 fun get_syn_decls thyname =
  let val (ref tab) = syntax_tab
  in
   case (Symtab.lookup tab thyname) of
     SOME decls => decls
   | NONE => []
  end  



(* *********** Translations ************ *)


 val trrule_tab = ref Symtab.empty : ((((xstring * string) Syntax.trrule) list) Symtab.table) ref

 fun add_trrule (thyname, trrule) = 
  let val (ref tab) = trrule_tab
      val trrules = (case (Symtab.lookup tab thyname) of
                      SOME l => l
                    | NONE => [])
      val tab' = Symtab.update (thyname, trrules @ [trrule]) tab
  in
   trrule_tab := tab'
  end 

 fun get_trrules thyname =
  let val (ref tab) = trrule_tab
  in
   case (Symtab.lookup tab thyname) of
     SOME trrules => trrules
   | NONE => []
  end  

 
(* *********** Translation functions ************ *)


 val tab_list = ref [parse_ast_tab, parse_tab, print_tab, print_ast_tab]
 fun index l 1 = hd l
   | index l n = index (tl l) (n-1)
 fun upd l t 1 = (let val t_ref = hd l
                      val _ = (t_ref := t)
                  in
                   t_ref::(tl l)
                  end)
   | upd l t n = (hd l)::(upd (tl l) t (n-1)) 

 fun add_trfun i (thyname, trfun) = 
  let val (ref tablist) = tab_list 
      val (ref tab) = index tablist i
      val trfuns = (case (Symtab.lookup tab thyname) of
                      SOME l => l
                    | NONE => [])
      val tab' = Symtab.update (thyname, trfuns @ [trfun]) tab
  in
   tab_list := (upd tablist tab' i)
  end 

 fun get_trfuns i thyname =
  let val (ref tablist) = tab_list
      val (ref tab) = index tablist i
  in
   case (Symtab.lookup tab thyname) of
     SOME trfuns => trfuns
   | NONE => []
  end  

 val add_parse_ast_trans = add_trfun 1
 val add_parse_trans = add_trfun 2  
 val add_print_trans = add_trfun 3
 val add_print_ast_trans = add_trfun 4
 fun add_token_trans (thyname, t) = 
  let val (ref tab) = token_tab
      val ts = (case (Symtab.lookup tab thyname) of
                      SOME l => l
                    | NONE => [])
      val tab' = Symtab.update (thyname, ts @ [t]) tab
  in
   token_tab := tab'
  end 

 val get_parse_ast_trans = get_trfuns 1
 val get_parse_trans = get_trfuns 2  
 val get_print_trans = get_trfuns 3
 val get_print_ast_trans = get_trfuns 4
 fun get_token_trans thyname =
  let val (ref tab) = token_tab
  in
   case (Symtab.lookup tab thyname) of
     SOME ts => ts
   | NONE => []
  end  



end